#include "IGL.h"
#include "Structures.h"
#include "Globals.h"
#include "DirectoryServices.h"

#include <math.h>

#include <list>
#include <vector>
#include <map>
#include <fstream>
using namespace std;


static inline double sqr(double x) {
    return x * x;
}


const int BACKGROUNDS = 20;


bool demoMode = false;


map<char, char> keyTranslation;
map<char, bool> keyState;

vector< pair<char*, pair<int, int> > > highScores; // Name, Difficulty, Score
int scoreDifficultyDisplay = 4, timeTillScoreChange = 0;


void loadHighScores(void) {
    char *buffer;
    int length, score, diff;
    char fullPath[PATH_MAX];
    
    strcpy(fullPath, getenv("HOME"));
    strcat(fullPath, "/.spaceinvaders");
        
    ifstream input(fullPath);
    
    highScores.clear();
    
    while (input) {
        input.read((char*)&length, sizeof(int));
        
        if (!input) {
            break;
        } else {
            if (length < 0) {
                fprintf(stderr, "Warning: high scores file is corrupt (invalid name length, %d).\n", length);
                break;
            } else {
                buffer = new char[length + 1];
                input.read(buffer, length);
                buffer[length] = '\0';
                
                input.read((char*)&score, sizeof(int));
                
                if (score < 0) {
                    fprintf(stderr, "Warning: high scores file is corrupt (invalid score, %d).\n", score);
                    break;
                } else {
                    input.read((char*)&diff, sizeof(int));
                    
                    if ((diff < 0) || (diff > 4)) {
                        fprintf(stderr, "Warning: high scores file is corrupt (invalid difficulty level, %d).\n", diff);
                    }
                    
                    highScores.push_back(pair<char*, pair<int, int> >(buffer, pair<int, int>(diff, score)));
                }
            }
        }
    }
    
    input.close();
}

bool addHighScore(const char *name, int difficulty, int score) {
    vector< pair<char*, pair<int, int> > >::iterator iter, stopper;
    int count = 0;
    char *buffer;
    
    for (iter = highScores.begin(), stopper = highScores.end(); iter != stopper; ++iter) {
        if ((*iter).second.first == difficulty) {
            ++count;
            
            if ((*iter).second.second < score) {
                if (10 == count) {
                    if (name) {
                        iter = highScores.erase(iter);
                    } else {
                        return true;
                    }
                }
                
                break;
            } else if (count >= 10) {
                return false;
            }
        }
    }
    
    if (name) {
        buffer = new char[strlen(name) + 1];
        strcpy(buffer, name);
        
        highScores.insert(iter, pair<char*, pair<int, int> >(buffer, pair<int, int>(difficulty, score)));
    }
    
    return true;
}

void saveHighScores(void) {
    vector< pair<char*, pair<int, int> > >::const_iterator iter, stopper;
    int length;
    char fullPath[PATH_MAX];
    
    strcpy(fullPath, getenv("HOME"));
    strcat(fullPath, "/.spaceinvaders");
    
    int output = open(fullPath, O_WRONLY | O_CREAT | O_TRUNC, 0644);
    
    if (output < 0) {
        fprintf(stderr, "Unable to save high scores, error #%d.\n", errno);
    } else {
        for (iter = highScores.begin(), stopper = highScores.end(); iter != stopper; ++iter) {
            length = strlen((*iter).first);
            
            write(output, (const char*)&length, sizeof(int));
            write(output, (*iter).first, length);
            write(output, (const char*)&((*iter).second.second), sizeof(int));
            write(output, (const char*)&((*iter).second.first), sizeof(int));
        }
        
        close(output);
    }
}


void doExit(int error) {
    //saveHighScores(); // Not needed anymore now that atexit() works
    exit(error);
}


const char PLAYER_ONE_UP = 'w';
const char PLAYER_ONE_LEFT = 'a';
const char PLAYER_ONE_DOWN = 's';
const char PLAYER_ONE_RIGHT = 'd';
const char PLAYER_ONE_SHOOT = '\t';

const char PLAYER_TWO_UP = 'i';
const char PLAYER_TWO_LEFT = 'j';
const char PLAYER_TWO_DOWN = 'k';
const char PLAYER_TWO_RIGHT = 'l';
const char PLAYER_TWO_SHOOT = ' ';


void setupDefaultKeyTranslations(void) {
    // Player One
    
    keyTranslation['w'] = PLAYER_ONE_UP;
    keyTranslation['a'] = PLAYER_ONE_LEFT;
    keyTranslation['s'] = PLAYER_ONE_DOWN;
    keyTranslation['d'] = PLAYER_ONE_RIGHT;
    keyTranslation['\t'] = PLAYER_ONE_SHOOT;
    
    // Player Two
    
    keyTranslation['i'] = PLAYER_TWO_UP;
    keyTranslation['j'] = PLAYER_TWO_LEFT;
    keyTranslation['k'] = PLAYER_TWO_DOWN;
    keyTranslation['l'] = PLAYER_TWO_RIGHT;
    keyTranslation[' '] = PLAYER_TWO_SHOOT;
}

Level currentLevel;
int currentLevelIndex = -1;


enum {TitleScreen, Initialising, MenuMain, MenuPlayers, MenuDifficulty, MenuOptions, MenuScores, MenuCheat, StartLevel, Running, Paused, FinishLevel, GameOver} GameState;
bool proceedToNextState = false;
int cheatState = 0;
bool runningFullScreen = false;


const int INPUT_BUFFER_SIZE = 50;
char inputBuffer[INPUT_BUFFER_SIZE];
int inputOffset = 0;
int highScorePlayer = -1;
bool askingForHighScore = false;


const int TINY_EXPLOSION = 50;
const int SMALL_EXPLOSION = 250;
const int MEDIUM_EXPLOSION = 1000;
const int BIG_EXPLOSION = 2500;
const int HUGE_EXPLOSION = 10000;
list<FadingPoint> points;

const int STAR_COUNT = 1000;
TypePoint stars[STAR_COUNT];
bool haveSetupStars = false;

const double updatePeriod = 0.04;

map<int, Shot*> shots;

EnemyInstance *enemies = NULL;

int playerCount = 0;
Player *players = NULL;

static inline EnemyInstance& enemyInstanceAt(int row, int column) {
    return enemies[(row * currentLevel.columnCount) + column];
}

TypePoint gridTopLeft;

static inline TypePoint enemyPosition(int row, int column) {
    EnemyInstance &temp = enemyInstanceAt(row, column);
    
    if (temp.inGrid) {
        return gridTopLeft + TypePoint(currentLevel.horizontalSpacing * (column + 0.5), -currentLevel.verticalSpacing * (row + 0.5));
    } else {
        return temp.position;
    }
}

bool gridMovingDown, gridMovingRight;
double desiredTop;
vector< pair<pair<TypePoint, TypePoint>, int> > alienShots; // First contains the current co-ordinate & velocity of the shot, second the type
vector< pair<TypeBMPPointer, TypePoint> > backgrounds; // Background sprites
vector<TypeBMPPointer> freeBackgrounds; // Backgrounds not yet in use

const double hudPowerWidth = 200.0;
const double hudHealthWidth = 200.0;


void resetPlayer(Player &player) {
    double difficultyModifier = 3.0 / (player.difficulty + 1);
    
    player.position = player.destination = TypePoint(xMax() / 2.0, 50);
    
    player.speed = (250 * updatePeriod) * difficultyModifier;
    player.rechargeRate = updatePeriod * difficultyModifier;
    player.currentCharge = 1.0;
    player.currentCapacity = 5.0 * difficultyModifier;
    player.waitCount = player.currentShotType = 0;
    
    player.invulnerabilityTime = (int)round((5 - player.difficulty) * (1.0 / updatePeriod));
}

void setupPlayer(Player &player, int difficulty) {
    player.difficulty = min(4, max(0, difficulty)); // Normalise the difficulty
    
    player.name = NULL; // Until we have something better to set it to
    
    changeToRelativeDirectory(graphicsDir);
    
    player.graphic.DeleteAll();
    
    if (!player.graphic.LoadFrom(playerGraphics)) {
        fprintf(stderr, "Unable to load player \"%s\".\n", playerGraphics);
        doExit(-26);
    }
    
    changeToRelativeDirectory(audioDir);
    
    if (!(player.dieSound = loadSoundFile(playerDieSound))) {
        fprintf(stderr, "Unable to load player die sound.\n");
        doExit(-27);
    }
    
    player.firing = false;
    player.didFire = true;
    player.score = 0;
    player.lives = 5 - player.difficulty;
    
    player.shots.clear();
    
    resetPlayer(player);
}


void setupStars(void) {
    int i;
    double xmax = xMax(), ymax = yMax();
    
    //printf("xmax = %f\n", xmax);
    
    for (i = 0; i < STAR_COUNT; ++i) {
        stars[i].x = Random(0.0, xmax);
        stars[i].y = Random(0.0, ymax);
    }
    
    haveSetupStars = true;
}

void drawHUD(void) {
    double xmax = xMax(), ymax = yMax();
    int playerIndex = 0;
    TypePolyList shipCopy;
    TypePoint shipSizeUL, shipSizeLR;
    double shipWidth;
    TypePoint lifePoint;
    char buffer[50];
    Shot *shot;
    
    for (playerIndex = 0; playerIndex < playerCount; ++playerIndex, ymax -= 20) {
        snprintf(buffer, 50, "Score: %d", players[playerIndex].score);
        
        TextColor = RGBA(0, 255, 0, 255);
        TextSize = 12.0;
        DrawTextAt(TypePoint(5, ymax - 15), buffer);
        
        if (players[playerIndex].lives > 0) {
            shipCopy = *(players[playerIndex].graphic.currentPoly());
            shot = shots[players[playerIndex].currentShotType];
            
            lifePoint.x = 110;
            lifePoint.y = ymax - 10;
            
            shipCopy.TransformByMatrix(ScaleMatrix(0.5, 0.5));
            
            shipCopy.boundingBox(shipSizeUL, shipSizeLR);
            shipWidth = shipSizeLR.x - shipSizeUL.x;
            
            for (int i = 1; i < players[playerIndex].lives; ++i) {
                shipCopy.CentreOn(lifePoint);
                shipCopy.Display();
                
                lifePoint.x += shipWidth + 5;
            }
            
            FillColor = RectangleColor = RGBA(255, 150, 150, 200);
            FillRectangle(xmax - (hudPowerWidth + 5), ymax - 5, xmax - (hudPowerWidth + 5) + (hudPowerWidth * (fmin(shot->cost, players[playerIndex].currentCharge) / players[playerIndex].currentCapacity)), ymax - 15);
            
            FillColor = RectangleColor = RGBA(150, 255, 150, 200);
            FillRectangle(xmax - (hudPowerWidth + 5) + (hudPowerWidth * (shot->cost / players[playerIndex].currentCapacity)), ymax - 5, xmax - (hudPowerWidth + 5) + (hudPowerWidth * (fmax(shot->cost, players[playerIndex].currentCharge) / players[playerIndex].currentCapacity)), ymax - 15);
            
            RectangleColor = RGBA(150, 255, 150, 255);
            DrawRectangle(xmax - (hudPowerWidth + 5), ymax - 5, xmax - 5, ymax - 15);
        }
    }
}

void spawnPoints(int numberOfPoints, const TypePoint &point, double minSpeed, double maxSpeed, long colour, double minLife, double maxLife) {
    TypePoint velocity;
    double speed, life;
    
    while (numberOfPoints--) {
        speed = minSpeed + ((1.0 - sqrt(Random(0.0, 1.0))) * (maxSpeed - minSpeed));
        
        velocity.x = -speed + (2 * Random(0.0, speed));
        velocity.y = sqrt(sqr(speed) - sqr(velocity.x));
        
        if (Random(0.0, 1.0) < 0.5) {
            velocity.y = -velocity.y;
        }
                
        life = Random(minLife, maxLife);
        
        points.push_back((FadingPoint){point, velocity, life, life, colour});
    }
}

void spawnPointsForPolyList(const TypePolyList &poly, int numberOfPoints, double minSpeed, double maxSpeed, double minLife, double maxLife) {
    int polyCount = poly.NrPolys();
    TypePoint point = poly.Centroid();
    
    for (int i = 0; i < polyCount; ++i) {
        spawnPoints((int)round((double)numberOfPoints / polyCount), point, minSpeed, maxSpeed, poly.GetPolyFillColor(i), minLife, maxLife);
    }
}

void paint(void) {
    EraseGraphics(BLACK);

    if (TitleScreen != GameState) {
        double xmax = xMax(), ymax = yMax();
        vector< pair<TypeBMPPointer, TypePoint> >::const_iterator iter, stopper;        
        
        for (iter = backgrounds.begin(), stopper = backgrounds.end(); iter != stopper; ++iter) {
            ShowBMPAt((*iter).first, (*iter).second);
        }
        
        PixelColor = RGBA(200, 0, 0, 228);
        DrawPixelsAt(stars, STAR_COUNT / 4);
        PixelColor = RGBA(150, 0, 150, 228);
        DrawPixelsAt(stars + (STAR_COUNT / 4), STAR_COUNT / 4);
        PixelColor = RGBA(0, 0, 200, 228);
        DrawPixelsAt(stars + (STAR_COUNT / 2), STAR_COUNT / 4);
        PixelColor = RGBA(100, 100, 100, 228);
        DrawPixelsAt(stars + (3 * (STAR_COUNT / 4)), STAR_COUNT / 4);
        
        if (MenuMain == GameState) {
            double centrePoint = xmax / 2.0;
            
            TextColor = RGBA(0, 255, 0, 255);
            TextSize = 28.0;
            
            DrawTextCentredAt(TypePoint(centrePoint, ymax - 50), "S P A C E   I N V A D E R S");
            
            TextSize = 18.0;
            DrawTextCentredAt(TypePoint(centrePoint, ymax - 70), "Main Menu");
            
            TextSize = 16.0;
            DrawTextCentredAt(TypePoint(centrePoint, ymax - 100), "1) New Game");
            DrawTextCentredAt(TypePoint(centrePoint, ymax - 120), "2) Options");
            DrawTextCentredAt(TypePoint(centrePoint, ymax - 140), "3) High Scores");
            DrawTextCentredAt(TypePoint(centrePoint, ymax - 180), "4) Quit");
        } else if (MenuOptions == GameState) {
            double centrePoint = xmax / 2.0;
            
            TextColor = RGBA(0, 255, 0, 255);
            TextSize = 28.0;
            
            DrawTextCentredAt(TypePoint(centrePoint, ymax - 50), "S P A C E   I N V A D E R S");
            
            TextSize = 18.0;
            DrawTextCentredAt(TypePoint(centrePoint, ymax - 70), "Options");
            
            TextSize = 16.0;
            DrawTextCentredAt(TypePoint(centrePoint, ymax - 100), "0) Return to Main Menu");
            
            if (runningFullScreen) {
                DrawTextCentredAt(TypePoint(centrePoint, ymax - 140), "1) Run Windowed");
            } else {
                DrawTextCentredAt(TypePoint(centrePoint, ymax - 140), "1) Run Fullscreen");
            }
        } else if (MenuPlayers == GameState) {
            double centrePoint = xmax / 2.0;
            
            TextColor = RGBA(0, 255, 0, 255);
            TextSize = 28.0;
            
            DrawTextCentredAt(TypePoint(centrePoint, ymax - 50), "N E W   G A M E");
            
            TextSize = 18.0;
            DrawTextCentredAt(TypePoint(centrePoint, ymax - 70), "Number of Players");
            
            TextSize = 16.0;
            
            if (demoMode) {
                DrawTextCentredAt(TypePoint(centrePoint, ymax - 100), "0) Turn Demo Mode Off");
            } else {
                DrawTextCentredAt(TypePoint(centrePoint, ymax - 100), "0) Turn Demo Mode On");
            }
            
            DrawTextCentredAt(TypePoint(centrePoint, ymax - 140), "1) One");
            DrawTextCentredAt(TypePoint(centrePoint, ymax - 160), "2) Two");
        } else if (MenuDifficulty == GameState) {
            double centrePoint = xmax / 2.0;
            
            TextColor = RGBA(0, 255, 0, 255);
            TextSize = 28.0;
            
            DrawTextCentredAt(TypePoint(centrePoint, ymax - 50), "N E W   G A M E");
            
            TextSize = 18.0;
            DrawTextCentredAt(TypePoint(centrePoint, ymax - 70), "Difficulty");
            
            TextSize = 16.0;
            DrawTextCentredAt(TypePoint(centrePoint, ymax - 100), "0) Blood Makes Me Dizzy");
            DrawTextCentredAt(TypePoint(centrePoint, ymax - 120), "1) Kind To Small Animals");
            DrawTextCentredAt(TypePoint(centrePoint, ymax - 140), "2) Tough & Ready");
            DrawTextCentredAt(TypePoint(centrePoint, ymax - 160), "3) Psychotic");
            DrawTextCentredAt(TypePoint(centrePoint, ymax - 180), "4) Ace Rimmer");
        } else if (MenuScores == GameState) {
            double centrePoint = xmax / 2.0;
            double position = ymax - 100;
            vector< pair<char*, pair<int, int> > >::const_iterator iter, stopper;
            char buffer[20];
            
            TextColor = RGBA(0, 255, 0, 255);
            TextSize = 28.0;
            
            DrawTextCentredAt(TypePoint(centrePoint, ymax - 50), "H I G H   S C O R E S");
            
            TextSize = 18.0;
            snprintf(buffer, 20, "Difficulty: %d", scoreDifficultyDisplay);
            DrawTextCentredAt(TypePoint(centrePoint, ymax - 70), buffer);
            
            TextSize = 16.0;
            
            for (iter = highScores.begin(), stopper = highScores.end(); iter != stopper; ++iter) {
                if ((*iter).second.first == scoreDifficultyDisplay) {
                    DrawTextAt(TypePoint(50, position), (*iter).first);
                    snprintf(buffer, 20, "%d", (*iter).second.second);
                    DrawTextLeftOf(TypePoint(xmax - 50, position), buffer);
                    
                    position -= 20;
                }
            }
        } else if (MenuCheat == GameState) {
            double centrePoint = xmax / 2.0;
            
            TextColor = RGBA(0, 255, 0, 255);
            TextSize = 30.0;
            
            DrawTextCentredAt(TypePoint(centrePoint, ymax - 50), "S M E G");
            
            TextSize = 16.0;
            DrawTextCentredAt(TypePoint(centrePoint, ymax - 100), "1) God Mode");
            DrawTextCentredAt(TypePoint(centrePoint, ymax - 120), "2) More Lives");
            DrawTextCentredAt(TypePoint(centrePoint, ymax - 140), "3) Skip One Level");
            DrawTextCentredAt(TypePoint(centrePoint, ymax - 160), "4) Skip Three Levels");
            DrawTextCentredAt(TypePoint(centrePoint, ymax - 180), "5) Skip Five Levels");
            DrawTextCentredAt(TypePoint(centrePoint, ymax - 200), "6) Increase Capacity");
            DrawTextCentredAt(TypePoint(centrePoint, ymax - 220), "7) Increase Recharge Rate");
        } else if (StartLevel == GameState) {        
            TextColor = RGBA(0, 255, 0, 255);
            TextSize = 30.0;
            
            DrawTextAtCentreTop(currentLevel.name);
            
            TextSize = 20.0;
            DrawTextAt(TypePoint(50, ymax - 75), currentLevel.introText, xmax - 100);
        } else if (FinishLevel == GameState) {
            TextColor = RGBA(0, 255, 0, 255);
            TextSize = 20.0;
            DrawTextAt(TypePoint(50, ymax - 50), currentLevel.endText, xmax - 100);
        } else if (Paused == GameState) {
            TextColor = RGBA(255, 255, 255, 255);
            TextSize = 20.0;
            DrawTextCentredAt(CentrePoint(), "PAUSED");
            
            TextSize = 14.0;
            DrawTextCentredAt(CentrePoint() - TypePoint(0.0, 25.0), "Press Escape to resume or Q to exit");
        } else if (GameOver == GameState) {
            TypePoint position = CentrePoint();
            char buffer[40];
            
            TextColor = RGBA(255, 255, 255, 255);
            TextSize = 40.0;
            position.y += 20.0;
            DrawTextCentredAt(position, "G A M E   O V E R");
            
            TextSize = 20.0;
            position.y -= 50.0;
            
            if ((highScorePlayer >= 0) && askingForHighScore) {
                snprintf(buffer, 40, "Enter your name, player %d:", highScorePlayer + 1);
                DrawTextCentredAt(position, buffer);
                
                position.y -= 30.0;
                DrawTextCentredAt(position, inputBuffer);
            } else if (highScorePlayer == -2) {
                DrawTextCentredAt(position, "(Press Any Key)");
            }
        } else if (Running == GameState) {
            int i, j;
            bool didDrawEnemy = false;
            vector< pair< pair<TypePoint, TypePoint>, int> >::iterator shotIter, shotStopper;
            Shot *shot;
            list<FadingPoint>::iterator pointIter, pointStopper;
            double rotationAngle;
            TypePoint centre;
            
            if (demoMode) {
                TextColor = RGBA(255, 255, 255, 200);
                TextSize = 20.0;
                DrawTextAtCentreTop("D E M O   M O D E");
                
                TextSize = 16.0;
                DrawTextCentredAt(TypePoint(xmax / 2.0, ymax - 50), "Press Escape to Exit");
            }
            
            for (shotIter = alienShots.begin(), shotStopper = alienShots.end(); shotIter != shotStopper; ++shotIter) {
                shot = shots[(*shotIter).second];
                
                centre = (*shotIter).first.first;
                
                rotationAngle = RadianAngleOfSlope((*shotIter).first.second.x, (*shotIter).first.second.y) - (PI / 2.0);
                
                shot->graphic.CentreOn(centre);
                shot->graphic.TransformByMatrix(RotationMatrixAbout(centre, rotationAngle));
                shot->graphic.Display();
                shot->graphic.TransformByMatrix(RotationMatrixAbout(centre, -rotationAngle));
            }
            
            for (j = 0; j < currentLevel.rowCount; ++j) {
                for (i = 0; i < currentLevel.columnCount; ++i) {
                    if (enemyInstanceAt(j, i).hitPoints > 0) {
                        didDrawEnemy = true;
                        
                        currentLevel.types[j].graphic.CentreOn(enemyPosition(j, i));
                        currentLevel.types[j].graphic.Display();
                    }
                }
            }
            
            for (pointIter = points.begin(), pointStopper = points.end(); pointIter != pointStopper; ++pointIter) {
                PixelColor = ((*pointIter).baseColour &0xFFFFFF00) | ((int)round(((*pointIter).life / (*pointIter).initialLife) * 255.0) & 0xFF);
                DrawPixelAt((*pointIter).position);
                
                (*pointIter).life -= updatePeriod;
                
                if ((*pointIter).life <= 0.0) {
                    pointIter = points.erase(pointIter);
                } else {
                    (*pointIter).position += (*pointIter).velocity;
                    (*pointIter).velocity.y += updatePeriod;
                }
            }
            
            if (!didDrawEnemy) {
                GameState = FinishLevel;
            } else {
                for (i = 0; i < playerCount; ++i) {
                    for (shotIter = players[i].shots.begin(), shotStopper = players[i].shots.end(); shotIter != shotStopper; ++shotIter) {
                        shot = shots[(*shotIter).second];
                        
                        centre = (*shotIter).first.first;
                        
                        rotationAngle = RadianAngleOfSlope((*shotIter).first.second.x, (*shotIter).first.second.y) - (PI / 2.0);
                        
                        shot->graphic.CentreOn(centre);
                        shot->graphic.TransformByMatrix(RotationMatrixAbout(centre, rotationAngle));
                        shot->graphic.Display();
                        shot->graphic.TransformByMatrix(RotationMatrixAbout(centre, -rotationAngle));
                    }
                    
                    if (players[i].lives > 0) {
                        players[i].graphic.CentreOn(players[i].position);
                        players[i].graphic.Display();
                        
                        if (players[i].destination != players[i].position) {
                            DrawMarkerAt(players[i].destination);
                        }
                    }
                }
                
                drawHUD();
            }
        }
    }
}

void mouseMoved(double x, double y) {
    if ((playerCount > 0) && !demoMode) {
        players[0].destination = TypePoint(x, y);
    }
}

bool performLoadLevel(void) {
    int i, j, shotType;
    Shot *shot;
        
    if (!loadLevel(currentLevelIndex, currentLevel)) {
        return false;
    }
    
    for (i = 0; i < playerCount; ++i) {
        players[i].position.y = 50;
    }
    
    for (i = 0; i < currentLevel.rowCount; ++i) {
        for (j = 0; j < currentLevel.types[i].numberOfShotTypes; ++j) {
            shotType = currentLevel.types[i].shotTypes[j];
            
            if (shots.end() == shots.find(shotType)) {
                shot = new Shot;
                
                if (!loadShot(shotType, *shot)) {
                    fprintf(stderr, "performLoadLevel() - Unable to load shot %d.\n", shotType);
                    return false;
                }
                
                shots[shotType] = shot;
            }
        }
    }
    
    return true;
}

void keyDownFunction(char key) {
    if (demoMode && (MenuPlayers != GameState) && (MenuDifficulty != GameState)) {
        if (ESCAPE == key) {
            demoMode = false;
            GameState = MenuMain;
        }
    } else {
        map<char, char>::iterator iter = keyTranslation.find(key);
        
        if (iter != keyTranslation.end()) {
            keyState[(*iter).second] = true;
        }
        
        switch (GameState) {
            case TitleScreen:
                ClearBackdrop();
                GameState = MenuMain;
                break;
            case MenuMain:
                switch (key) {
                    case '1':
                        if (players) {
                            for (int i = 0; i < playerCount; ++i) {
                                players[i].graphic.DeleteAll();
                            }
                            
                            free(players);
                            
                            players = NULL;
                        }
                        
                        playerCount = 0;
                        
                        GameState = MenuPlayers;
                        break;
                    case '2':
                        GameState = MenuOptions;
                        break;
                    case '3':
                        GameState = MenuScores;
                        break;
                    case '4':
                        doExit(0);
                        break;
                }
                
                break;
            case MenuOptions:
                switch (key) {
                    case '0':
                        GameState = MenuMain;
                        break;
                    case '1':
                        if (runningFullScreen) {
                            RunWindowed();
                        } else {
                            RunFullScreen();
                        }
                        
                        runningFullScreen ^= 1;
                        
                        break;
                    default:
                        break;
                }
                
                break;
            case MenuPlayers:
                if (players) {
                    delete[] players;
                    players = NULL;
                }
                
                switch (key) {
                    case '0':
                        demoMode ^= 1;
                        break;
                    case '1':
                        playerCount = 1;
                        //players = (Player*)malloc(sizeof(Player));
                        players = new Player[1];
                        GameState = MenuDifficulty;
                        break;
                    case '2':
                        playerCount = 2;
                        //players = (Player*)malloc(sizeof(Player) * 2);
                        players = new Player[2];
                        GameState = MenuDifficulty;
                        break;
                }
                
                break;
            case MenuDifficulty:
                if ((key >= '0') && (key <= '4')) {
                    for (int i = 0; i < playerCount; ++i) {
                        setupPlayer(players[i], key - '0');
                    }
                    
                    if ('4' == key) {
                        changeToRelativeDirectory(audioDir);
                        PlayWAVFile("Ace.mp3");
                    }
                    
                    if (currentLevelIndex >= 0) {
                        unloadLevel(currentLevel);
                    }
                    
                    currentLevelIndex = 0;
                    
                    if (!performLoadLevel()) {
                        fprintf(stderr, "Unable to load first level.\n");
                        doExit(-18);
                    }
                    
                    GameState = Initialising;
                }
                
                break;
            case MenuScores:
                switch (key) {
                    case LEFT_ARROW:
                        scoreDifficultyDisplay = (scoreDifficultyDisplay + 4) % 5;
                        timeTillScoreChange = (int)round(5.0 * (1.0 / updatePeriod));
                        break;
                    case RIGHT_ARROW:
                        scoreDifficultyDisplay = (scoreDifficultyDisplay + 1) % 5;
                        timeTillScoreChange = (int)round(5.0 * (1.0 / updatePeriod));
                        break;
                    default:
                        GameState = MenuMain;
                }
                
                break;
            case MenuCheat:
                int p;
                
                switch (key) {
                    case '1':
                        for (p = 0; p < playerCount; ++p) {
                            if (players[p].invulnerabilityTime > (1 << 28)) {
                                players[p].invulnerabilityTime = 0;
                            } else {
                                players[p].invulnerabilityTime = (1 << 30); // That's roughly 18 months of invulnerability time, if you're curious
                            }
                        }
                        
                        break;
                    case '2':
                        for (p = 0; p < playerCount; ++p) {
                            players[p].lives += 3;
                        }
                        
                        break;
                    case '3':
                        GameState = FinishLevel;
                        
                        break;
                    case '4':
                        currentLevelIndex += 2;
                        
                        GameState = FinishLevel;
                        
                        break;
                    case '5':
                        currentLevelIndex += 4;
                        
                        GameState = FinishLevel;
                        
                        break;
                    case '6':
                        for (p = 0; p < playerCount; ++p) {
                            players[p].currentCapacity *= 2;
                        }
                        
                        break;
                    case '7':
                        for (p = 0; p < playerCount; ++p) {
                            players[p].rechargeRate *= 2;
                        }
                        
                        break;
                    default:
                        GameState = Running;
                }
                    
                    break;
            case Running:
                switch (key) {
                    case ESCAPE:
                        GameState = Paused;
                        break;
                    case 's':
                        if (0 == cheatState) {
                            cheatState = 1; break;
                        }
                    case 'm':
                        if (1 == cheatState) {
                            cheatState = 2; break;
                        }
                    case 'e':
                        if (2 == cheatState) {
                            cheatState = 3; break;
                        }
                    case 'g':
                        if (3 == cheatState) {
                            GameState = MenuCheat;
                        }
                    default:
                        cheatState = 0;
                }
                
                // Do appropriate stuff here
                
                break;
            case Paused:
                switch (key) {
                    case ESCAPE:
                        GameState = Running;
                        break;
                    case 'q':
                    case 'Q':
                        doExit(0);
                }
                
                break;
            case StartLevel:                
                for (p = 0; p < playerCount; ++p) {
                    players[p].invulnerabilityTime = (int)round(fmax(players[p].invulnerabilityTime, ((int)round((5 - players[p].difficulty) * (1.0 / updatePeriod)))));
                }

                // Fall through
            case FinishLevel:
                switch (key) {
                    case RETURN:
                    case ENTER:
                    case SPACE_BAR:
                        proceedToNextState = true;
                        break;
                }
                
                break;
            case GameOver:
                if ((highScorePlayer >= 0) && askingForHighScore) {
                    if ((RETURN == key) || (ENTER == key)) {
                        if (inputOffset > 0) {
                            scoreDifficultyDisplay = players[highScorePlayer].difficulty;
                            assert(addHighScore(inputBuffer, scoreDifficultyDisplay, players[highScorePlayer].score));
                            askingForHighScore = false;
                            inputOffset = 0;
                            inputBuffer[0] = '\0';
                        }
                    } else if ((isalnum(key) || (SPACE_BAR == key)) && ((inputOffset + 1) < INPUT_BUFFER_SIZE)) {
                        inputBuffer[inputOffset++] = key;
                        inputBuffer[inputOffset] = '\0';
                    } else if ((inputOffset > 0) && (BACKSPACE == key)) {
                        --inputOffset;
                        inputBuffer[inputOffset] = '\0';
                    }
                } else if (highScorePlayer == -2) {
                    highScorePlayer = -1;
                    timeTillScoreChange = (int)round(5.0 * (1.0 / updatePeriod));
                    GameState = MenuScores;
                }
                
                break;
            default:
                break; // Do nothing
        }
    }
}

void keyUpFunction(char key) {
    if (!demoMode) {
        map<char, char>::iterator iter = keyTranslation.find(key);
        
        if (iter != keyTranslation.end()) {
            if (0 < playerCount) {
                if (PLAYER_ONE_SHOOT == (*iter).second) {
                    players[0].firing = false;
                } else if ((1 < playerCount) && (PLAYER_TWO_SHOOT == (*iter).second)) {
                    players[1].firing = false;
                }
            }
            
            keyState[(*iter).second] = false;
        }
    }
}

void buttonDown(int button, double x, double y) {
    if (!demoMode) {
        switch (GameState) {
            case Running:
                if (playerCount > 0) {
                    players[0].firing = true;
                    players[0].didFire = false;
                }
                
                break;
            case StartLevel:
            case FinishLevel:
                proceedToNextState = true;
                
                break;
            default:
                break; // Do nothing
        }
    }
}

void buttonUp(int button, double x, double y) {
    if (!demoMode) {
        switch (GameState) {
            case TitleScreen:
                ClearBackdrop();
                GameState = MenuMain;
                break;
            case Running:
                if (playerCount > 0) {
                    players[0].firing = false;
                }
                
                break;
            default:
                break; // Do nothing
        }
    }
}

void gridBoundingBox(TypePoint &ULCorner, TypePoint &LRCorner) {
    int i, j;
    TypePoint current;

    ULCorner.invalidate();
    LRCorner.invalidate();
    
    for (i = 0; i < currentLevel.columnCount; ++i) {
        for (j = 0; j < currentLevel.rowCount; ++j) {
            if ((enemyInstanceAt(j, i).hitPoints > 0) && (enemyInstanceAt(j, i).inGrid)) {
                current = enemyPosition(j, i);
                
                if (!ULCorner.valid()) {
                    ULCorner = LRCorner = current;
                } else {
                    if (current.x < ULCorner.x) {
                        ULCorner.x = current.x;
                    } else if (current.x > LRCorner.x) {
                        LRCorner.x = current.x;
                    }
                    
                    if (current.y > ULCorner.y) {
                        ULCorner.y = current.y;
                    } else if (current.y < LRCorner.y) {
                        LRCorner.y = current.y;
                    }
                }
            }
        }
    }
    
    ULCorner.x -= (currentLevel.horizontalSpacing / 2.0);
    LRCorner.x += (currentLevel.horizontalSpacing / 2.0);
    
    ULCorner.y += (currentLevel.verticalSpacing / 2.0);
    LRCorner.y -= (currentLevel.verticalSpacing / 2.0);
}

void updateAnimations(void) {
    int i;
    map<int, Shot*>::iterator iter, stopper;

    if (currentLevel.rowCount > 0) {
        for (i = 0; i < currentLevel.rowCount; ++i) {
            currentLevel.types[i].graphic.step();
        }
    }
    
    for (i = 0; i < playerCount; ++i) {
        players[i].graphic.step();
    }
    
    for (iter = shots.begin(), stopper = shots.end(); iter != stopper; ++iter) {
        (*iter).second->graphic.step();
    }
}

bool update(void) {
    vector< pair<TypeBMPPointer, TypePoint> >::iterator iter, stopper;
    static double backgroundClear = 0.0;
    int i, width, height;
    double xmax = xMax(), ymax = yMax();
    
    if (!haveSetupStars) {
        setupStars();
    } else {        
        for (i = 0; i < STAR_COUNT; ++i) {
            if (stars[i].y > ymax) {
                stars[i].x = Random(0.0, xmax);
                stars[i].y -= ymax;
            } else {
                stars[i].y += ((i % 100) + 1) / 25.0;
            }
        }
    }
    
    for (iter = backgrounds.begin(), stopper = backgrounds.end(); iter != stopper; ++iter) {
        getImageSize((*iter).first, width, height);
        
        if ((*iter).second.y - height > ymax) {
            freeBackgrounds.push_back((*iter).first);
            iter = backgrounds.erase(iter);
        } else {
            (*iter).second.y += 0.05;
        }
    }
    
    backgroundClear -= 0.05;
    
    if ((backgroundClear < 0.0) && !freeBackgrounds.empty() && (Random(0.0, 1.0) < 0.1)) {
        int selection = (int)floor(Random(0.0, freeBackgrounds.size()));
        vector<TypeBMPPointer>::iterator iter2 = freeBackgrounds.begin();
        
        while (selection-- > 0) {
            ++iter2;
        }
        
        getImageSize(*iter2, width, height);
        
        backgroundClear = height;
        
        backgrounds.push_back(pair<TypeBMPPointer, TypePoint>(*iter2, TypePoint(Random(-width, xmax), 0.0)));
        freeBackgrounds.erase(iter2);
    }
    
    if (Initialising == GameState) {
        int j;
        TypePoint currentPos;
        
        cheatState = 0;
        inputOffset = 0;
        highScorePlayer = -1;
        askingForHighScore = false;
        
        gridTopLeft.x = 0.0;
        gridTopLeft.y = ymax;
        
        currentPos.y = gridTopLeft.y + (currentLevel.verticalSpacing / 2.0);
        
        if (enemies) {
            free(enemies);
        }
        
        enemies = (EnemyInstance*)malloc(sizeof(EnemyInstance) * currentLevel.rowCount * currentLevel.columnCount);
        
        for (j = 0; j < currentLevel.rowCount; ++j) {
            currentPos.x = gridTopLeft.x - (currentLevel.horizontalSpacing / 2.0);
            currentPos.y -= currentLevel.verticalSpacing;
            
            for (i = 0; i < currentLevel.columnCount; ++i) {
                currentPos.x += currentLevel.horizontalSpacing;
                
                enemyInstanceAt(j, i) = (EnemyInstance){true, currentPos, currentLevel.types[j].hitPoints};
            }
        }
        
        gridMovingDown = false;
        gridMovingRight = true;
        
        GameState = StartLevel;
        proceedToNextState = false;
    } else if (StartLevel == GameState) {
        if (currentLevel.backgroundMusic) {
            playSound(currentLevel.backgroundMusic, true);
        }
        
        if (proceedToNextState || demoMode) {
            GameState = Running;
            proceedToNextState = false;
        }
    } else if (FinishLevel == GameState) {
        if (currentLevel.backgroundMusic) {
            stopPlaying(currentLevel.backgroundMusic, 3.0);
        }
        
        if (proceedToNextState || demoMode) {
            unloadLevel(currentLevel);
            
            ++currentLevelIndex;
            
            if (!performLoadLevel()) { // Presume end of game
                GameState = GameOver;
            } else {
                GameState = Initialising;
            }
            
            proceedToNextState = false;
        }
    } else if (MenuScores == GameState) {
        --timeTillScoreChange;
        
        if (timeTillScoreChange < 0) {
            scoreDifficultyDisplay = (scoreDifficultyDisplay + 1) % 5;
            timeTillScoreChange = (int)round(5.0 * (1.0 / updatePeriod));
        }
    } else if (Running == GameState) {
        int j;
        TypePoint bbUL, bbLR, bbUL2, bbLR2;
        vector< pair< pair<TypePoint, TypePoint>, int> >::iterator shotIter, shotStopper, shotIter2, shotStopper2;
        Shot *shot, *shot2;
        list<FadingPoint>::iterator pointIter, pointStopper;
        FadingPoint current;
        double xmin = xMin(), ymin = yMin();
        bool atLeastOnePlayerIsAlive = false;
        
        updateAnimations();
        
        if ((playerCount > 0) && !demoMode) {
            if (keyState[PLAYER_ONE_UP]) {
                players[0].destination.y += players[0].speed;
            }
            
            if (keyState[PLAYER_ONE_DOWN]) {
                players[0].destination.y -= players[0].speed;
            }
            
            if (keyState[PLAYER_ONE_LEFT]) {
                players[0].destination.x -= players[0].speed;
            }
            
            if (keyState[PLAYER_ONE_RIGHT]) {
                players[0].destination.x += players[0].speed;
            }
            
            if (keyState[PLAYER_ONE_SHOOT]) {
                players[0].firing = true;
                players[0].didFire = false;
            }
            
            if (playerCount > 1) {
                if (keyState[PLAYER_TWO_UP]) {
                    players[1].destination.y += players[1].speed;
                }
                
                if (keyState[PLAYER_TWO_DOWN]) {
                    players[1].destination.y -= players[1].speed;
                }
                
                if (keyState[PLAYER_TWO_LEFT]) {
                    players[1].destination.x -= players[1].speed;
                }
                
                if (keyState[PLAYER_TWO_RIGHT]) {
                    players[1].destination.x += players[1].speed;
                }
                
                if (keyState[PLAYER_TWO_SHOOT]) {
                    players[1].firing = true;
                    players[1].didFire = false;
                }
            }
        }
        
        for (pointIter = points.begin(), pointStopper = points.end(); pointIter != pointStopper; ++pointIter) {
            current = *pointIter;
            
            current.life -= updatePeriod;
            
            if (current.life <= 0.0) {
                pointIter = points.erase(pointIter);
            } else {
                current.position += current.velocity;
                current.velocity.y += updatePeriod * 2.0;
                
                if ((current.position.x < xmin) || (current.position.x > xmax) || (current.position.y < ymin) || (current.position.y > ymax)) {
                    pointIter = points.erase(pointIter);
                } else {
                    *pointIter = current;
                }
            }
        }
        
        for (j = 0; j < currentLevel.rowCount; ++j) {
            for (i = 0; i < currentLevel.columnCount; ++i) {
                if (enemyInstanceAt(j, i).hitPoints > 0) {
                    for (int k = 0; k < currentLevel.types[j].numberOfShotTypes; ++k) {
                        if (Random(0.0, 1.0) < ((currentLevel.types[j].shotProbabilities[k] * updatePeriod))) {
                            TypePoint velocity;
                            
                            shot = shots[currentLevel.types[j].shotTypes[k]];
                            playSound(shot->sound);
                            
                            switch (shot->behaviour) {
                                case Falling:
                                case Homing:
                                    velocity.x = 0.0;
                                    velocity.y = -shot->speed;
                                    break;
                                case Aimed:
                                    // To be completed
                                    break;
                                default:
                                    fprintf(stderr, "Unknown shot type.\n");
                            }
                            
                            alienShots.push_back(pair<pair<TypePoint, TypePoint>, int>(pair<TypePoint, TypePoint>(enemyPosition(j, i), velocity), currentLevel.types[j].shotTypes[k]));
                        }
                    }
                    
                    currentLevel.types[j].graphic.boundingBoxAtPoint(enemyPosition(j, i), bbUL, bbLR);
                    
                    for (int p = 0; p < playerCount; ++p) {
                        if (players[p].lives > 0) {
                            if (players[p].invulnerabilityTime <= 0) {
                                players[p].graphic.boundingBoxAtPoint(players[p].position, bbUL2, bbLR2);
                                
                                if (BoxesOverlap(bbUL, bbLR, bbUL2, bbLR2)) {
                                    enemyInstanceAt(j, i).hitPoints = 0;
                                    currentLevel.types[j].graphic.CentreOn(enemyPosition(j, i));
                                    playSound(currentLevel.types[j].dieSound);
                                    spawnPointsForPolyList(*(currentLevel.types[j].graphic.currentPoly()), MEDIUM_EXPLOSION, 0.0, 3.0, 1.0, 5.0);
                                    
                                    players[p].graphic.CentreOn(players[p].position);
                                    playSound(players[p].dieSound);
                                    spawnPointsForPolyList(*(players[p].graphic.currentPoly()), BIG_EXPLOSION, 0.0, 6.0, 5.0, 10.0);
                                    
                                    --(players[p].lives);
                                    
                                    resetPlayer(players[p]);
                                }
                            }
                        }
                        
                        for (shotIter = players[p].shots.begin(), shotStopper = players[p].shots.end(); shotIter != shotStopper; ++shotIter) {
                            shot = shots[(*shotIter).second];
                            
                            shot->graphic.boundingBoxAtPoint((*shotIter).first.first, bbUL2, bbLR2);
                            
                            if (BoxesOverlap(bbUL, bbLR, bbUL2, bbLR2)) {                                    
                                enemyInstanceAt(j, i).hitPoints -= shot->damage;
                                
                                if (enemyInstanceAt(j, i).hitPoints <= 0) {
                                    players[p].score += currentLevel.types[j].value;
                                    
                                    currentLevel.types[j].graphic.CentreOn(enemyPosition(j, i));
                                    playSound(currentLevel.types[j].dieSound);
                                    spawnPointsForPolyList(*(currentLevel.types[j].graphic.currentPoly()), MEDIUM_EXPLOSION, 0.0, 3.0, 1.0, 5.0);
                                } else {
                                    playSound(currentLevel.types[j].hitSound);
                                }
                                
                                shot->graphic.CentreOn((*shotIter).first.first);
                                spawnPointsForPolyList(*(shot->graphic.currentPoly()), SMALL_EXPLOSION, 0.0, 2.0, 0.5, 3.0);
                                
                                shotIter = players[p].shots.erase(shotIter);
                            }
                        }
                    }
                }
            }
        }
        
        for (shotIter = alienShots.begin(), shotStopper = alienShots.end(); shotIter != shotStopper; ++shotIter) {
            if ((*shotIter).first.first.y < ymin) {
                shotIter = alienShots.erase(shotIter);
            } else {
                TypePoint targetVector;

                shot = shots[(*shotIter).second];
                
                switch (shot->behaviour) {
                    case Falling:
                    case Aimed:
                        break; // Do nothing
                    case Homing:
                        double bestDistance, temp;
                        int choice;
                        double angle;
                        int p;
                        
                        bestDistance = FLT_MAX;
                        choice = -1;
                        
                        for (p = 0; p < playerCount; ++p) {
                            if (players[p].lives > 0) {
                                angle = RadianAngleBetween((*shotIter).first.second, players[p].position - (*shotIter).first.first);

                                if (angle < (PI / 8.0)) {
                                    temp = Distance(players[p].position, (*shotIter).first.second);
                                    
                                    if (temp < bestDistance) {
                                        bestDistance = temp;
                                        choice = p;
                                    }
                                }
                            }
                        }

                        if (choice >= 0) {
                            targetVector = players[choice].position - (*shotIter).first.first;
                                
                            (*shotIter).first.second += (shot->speed * Normalize(targetVector));
                        }
                        
                        (*shotIter).first.second = Normalize((*shotIter).first.second) * shot->speed;
                        
                        break;
                    default:
                        fprintf(stderr, "Unknown shot type.\n");
                }
                
                (*shotIter).first.first += (*shotIter).first.second;
                
                shot->graphic.boundingBoxAtPoint((*shotIter).first.first, bbUL, bbLR);
                
                if ((bbLR.x < xmin) || (bbUL.x > xmax) || (bbUL.y < ymin) || (bbLR.y > ymax)) {
                    shotIter = alienShots.erase(shotIter);
                } else {
                    for (int p = 0; p < playerCount; ++p) {
                        if (players[p].lives > 0) {
                            players[p].graphic.boundingBoxAtPoint(players[p].position, bbUL2, bbLR2);
                            
                            if (BoxesOverlap(bbUL, bbLR, bbUL2, bbLR2)) {                                    
                                shot->graphic.CentreOn((*shotIter).first.first);
                                spawnPointsForPolyList(*(shot->graphic.currentPoly()), SMALL_EXPLOSION, 0.0, 2.0, 0.5, 3.0);
                                
                                shotIter = alienShots.erase(shotIter);
                                
                                if (players[p].invulnerabilityTime <= 0) {
                                    players[p].graphic.CentreOn(players[p].position);
                                    playSound(players[p].dieSound);
                                    spawnPointsForPolyList(*(players[p].graphic.currentPoly()), BIG_EXPLOSION, 0.0, 6.0, 5.0, 10.0);
                                    
                                    --(players[p].lives);
                                    
                                    resetPlayer(players[p]);
                                }
                            }
                        }
                    }
                }
            }
        }
        
        for (int p = 0; p < playerCount; ++p) {
            if (demoMode && (players[p].destination == players[p].position)) {
                players[p].destination = TypePoint(Random(xmin, xmax), Random(ymin, ymax / 2.0));
            }
            
            for (shotIter = players[p].shots.begin(), shotStopper = players[p].shots.end(); shotIter != shotStopper; ++shotIter) {
                shot = shots[(*shotIter).second];
                
                switch (shot->behaviour) {
                    case Falling:
                    case Aimed:
                        break; // Do nothing
                    case Homing:
                        // To be completed
                        break;
                    default:
                        fprintf(stderr, "Unknown shot type.\n");
                }
                
                (*shotIter).first.first += (*shotIter).first.second;
                
                shot->graphic.boundingBoxAtPoint((*shotIter).first.first, bbUL, bbLR);
                
                if ((bbLR.x < xmin) || (bbUL.x > xmax) || (bbUL.y < ymin) || (bbLR.y > ymax)) {
                    shotIter = players[p].shots.erase(shotIter);
                } else {
                    for (shotIter2 = alienShots.begin(), shotStopper2 = alienShots.end(); shotIter2 != shotStopper2; ++shotIter2) {
                        shot2 = shots[(*shotIter2).second];
                        
                        shot2->graphic.boundingBoxAtPoint((*shotIter2).first.first, bbUL2, bbLR2);
                        
                        if (BoxesOverlap(bbUL, bbLR, bbUL2, bbLR2)) {                                    
                            shot->graphic.CentreOn((*shotIter).first.first);
                            spawnPointsForPolyList(*(shot->graphic.currentPoly()), SMALL_EXPLOSION, 0.0, 2.0, 0.5, 3.0);
                            
                            shot2->graphic.CentreOn((*shotIter2).first.first);
                            spawnPointsForPolyList(*(shot2->graphic.currentPoly()), SMALL_EXPLOSION, 0.0, 2.0, 0.5, 3.0);
                            
                            alienShots.erase(shotIter2);
                            shotIter = players[p].shots.erase(shotIter);
                            break;
                        }
                    }
                }
            }
            
            if (players[p].lives > 0) {
                atLeastOnePlayerIsAlive = true;
                
                --(players[p].invulnerabilityTime);

                players[p].currentCharge = fmin(players[p].currentCapacity, players[p].currentCharge + players[p].rechargeRate);
                --(players[p].waitCount);
                
                shot = shots[players[p].currentShotType];
                
                if ((demoMode || players[p].firing || !players[p].didFire) && (players[p].currentCharge >= shot->cost) && (players[p].waitCount <= 0)) {
                    TypePoint velocity;
                    
                    playSound(shot->sound);
                    
                    switch (shot->behaviour) {
                        case Falling:
                            velocity.x = 0.0;
                            velocity.y = shot->speed;
                            break;
                        case Aimed: // To be completed; make it actually aim
                            velocity.x = 0.0;
                            velocity.y = shot->speed;
                            break;
                        case Homing:
                            velocity.x = velocity.y = 0.0; // Initially stationary
                            break; 
                        default:
                            fprintf(stderr, "Unknown shot type.\n");
                    }
                    
                    players[p].shots.push_back(pair<pair<TypePoint, TypePoint>, int>(pair<TypePoint, TypePoint>(players[p].position, velocity), players[p].currentShotType));
                    
                    players[p].currentCharge -= shot->cost;
                    players[p].waitCount = shot->reloadTime;
                    
                    players[p].didFire = true;
                }
                
                players[p].graphic.CentreOn(players[p].destination);
                players[p].graphic.boundingBox(bbUL, bbLR);
                
                if (bbUL.x < xmin) {
                    players[p].destination.x += (xmin - bbUL.x);
                } else if (bbLR.x > xmax) {
                    players[p].destination.x -= (bbLR.x - xmax);
                }
                
                if (bbLR.y < ymin) {
                    players[p].destination.y += (ymin - bbLR.y);
                } else if (bbUL.y > ymax) {
                    players[p].destination.y -= (bbUL.y - ymax);
                }
                
                players[p].position = LerpPointByDistance(players[p].position, players[p].destination, fmin(Distance(players[p].position, players[p].destination), players[p].speed));
                
                players[p].graphic.CentreOn(players[p].position);
                players[p].graphic.boundingBox(bbUL, bbLR);
                
                if (bbUL.x < xmin) {
                    players[p].position.x += (xmin - bbUL.x);
                } else if (bbLR.x > xmax) {
                    players[p].position.x -= (bbLR.x - xmax);
                }
                
                if (bbLR.y < ymin) {
                    players[p].position.y += (ymin - bbLR.y);
                } else if (bbUL.y > ymax) {
                    players[p].position.y -= (bbUL.y - ymax);
                }
            }
        }
        
        if (!atLeastOnePlayerIsAlive) {
            GameState = GameOver;
        } else {
            gridBoundingBox(bbUL, bbLR);
            
            if (gridMovingDown) {                
                gridTopLeft.y -= currentLevel.speed;
                
                if (gridTopLeft.y <= desiredTop) {
                    gridTopLeft.y = desiredTop;
                    gridMovingDown = false;
                } else if (gridTopLeft.y < ymin) {
                    gridTopLeft.y += ymax + (bbUL.y - bbLR.y);
                }
            } else {
                if (gridMovingRight) {
                    gridTopLeft.x += currentLevel.speed;
                    
                    if ((bbLR.x + currentLevel.speed) > xmax) {
                        gridTopLeft.x -= ((bbLR.x + currentLevel.speed) - xmax);
                        gridMovingRight = false;
                        gridMovingDown = true;
                        
                        desiredTop = gridTopLeft.y - currentLevel.verticalSpacing;
                    }
                } else {
                    gridTopLeft.x -= currentLevel.speed;
                    
                    if ((bbUL.x - currentLevel.speed) < xmin) {
                        gridTopLeft.x += (xmin - (bbUL.x - currentLevel.speed));
                        
                        gridMovingRight = true;
                        gridMovingDown = true;
                        
                        desiredTop = gridTopLeft.y - currentLevel.verticalSpacing;
                    }
                }
            }
        }
    } else if (GameOver == GameState) {
        if (demoMode) {
            GameState = MenuScores;
        } else {
            if ((highScorePlayer >= -1) && !askingForHighScore) {
                for (int p = highScorePlayer + 1; p < playerCount; ++p) {
                    if (addHighScore(NULL, players[p].difficulty, players[p].score)) {
                        scoreDifficultyDisplay = players[p].difficulty;
                        highScorePlayer = p;
                        askingForHighScore = true;
                        break;
                    }
                }
                
                if (!askingForHighScore) {
                    highScorePlayer = -2;
                }
            }
        }
    }
    
    return true;
}

int main(int argc, char *argv[]) {
    char buffer[20];
    Shot *shot = new Shot;
    
    InitGraphics(argc, argv);
    
    initDirectoryServices();
    setupDefaultKeyTranslations();
    loadHighScores();
        
    atexit(saveHighScores);
    
    DefineUserCoordinates(0, 0, 1024, 768);
    glutReshapeWindow(512, 512);
    
    GameState = TitleScreen;
    changeToRelativeDirectory(graphicsDir);
    ShowBackdrop(titleScreenFileName);
    
    if (!loadShot(0, *shot)) { // Temporary, until weapons inventory is loaded into levels
        fprintf(stderr, "Unable to load shot 0.\n");
        doExit(-28);
    }
    
    shots[0] = shot;
    
    changeToRelativeDirectory(backgroundsDir);
    
    for (int i = 1; i <= BACKGROUNDS; ++i) {
        snprintf(buffer, 20, backgroundsFormat, i);
        freeBackgrounds.push_back(LoadBMPImage(buffer));
    }
    
    MarkerColor = RGBA(0, 0, 255, 90);
    
    SetPaintFunction(paint);
    SetKeyPressFunction(keyDownFunction);
    SetKeyReleaseFunction(keyUpFunction);
    SetButtonDownFunction(buttonDown);
    SetButtonUpFunction(buttonUp);
    SetMouseMovedFunction(mouseMoved);
    SetMouseDraggedFunction(mouseMoved);
    SetRegularTimerFunction(update, updatePeriod);
    
    gridTopLeft.x = gridTopLeft.y = -1.0;
    
    //RunFullScreen();
    
    RunGraphics();
    
    return 0;
}
