#include "Structures.h"


char* readQuotedString(istream &input) {
    static char *buffer = NULL;
    char temp, *result = NULL;
    static int bufferSize = 512;
    int offset = 0;
    
    if (!buffer) {
        bufferSize = 512;
        buffer = (char*)malloc(bufferSize);
    }
    
    input.ignore(1000000, '"');
    
    while (input && ((temp = input.get()) != '"')) {
        if (offset + 2 == bufferSize) {
            bufferSize += 512;
            buffer = (char*)realloc(buffer, bufferSize);
        }
        
        buffer[offset] = temp;
        ++offset;
    }
    
    if (input) {
        buffer[offset] = '\0';
        ++offset;
        
        result = (char*)malloc(offset);
        strncpy(result, buffer, offset);
        
        if (bufferSize > 4096) { // If we're using more than a page, we should shrink it... otherwise we can probably survive leaving it around
            free(buffer);
            buffer = NULL;
        }
    }
    
    return result;
}

bool loadShot(int type, Shot &shot) {
    char *temp;
    char buffer[20];
    
    changeToRelativeDirectory(shotsDir);
    
    snprintf(buffer, 20, "%d", type);
    ifstream input(buffer);
    
    if (!input) {
        fprintf(stderr, "Unable to find/open shot type %d.\n", type);
        
        return false;
    }
    
    shot.name = readQuotedString(input);
    
    if (!shot.name) {
        fprintf(stderr, "Unable to read shot name for type %d.\n", type);
        
        input.close();
        return false;
    }
    
    temp = readQuotedString(input);
    
    if (!temp) {
        fprintf(stderr, "Unable to read shot graphic name for type %d.\n", type);

        free(shot.name);
        input.close();
        return false;
    }
    
    changeToRelativeDirectory(graphicsDir);
    
    if (!shot.graphic.LoadFrom(temp)) {
        fprintf(stderr, "Unable to read shot graphic for type %d.\n", type);

        free(shot.name);
        free(temp);
        input.close();
        return false;
    }
    
    free(temp);
    
    temp = readQuotedString(input);
    
    if (!temp) {
        fprintf(stderr, "Unable to read shot sound name for type %d.\n", type);
        
        free(shot.name);
        input.close();
        return false;
    }
    
    changeToRelativeDirectory(audioDir);
    
    if (!(shot.sound = loadSoundFile(temp))) {
        fprintf(stderr, "Unable to read shot sound for type %d.\n", type);
        
        free(shot.name);
        free(temp);
        input.close();
        return false;
    }
    
    free(temp);
    
    input >> shot.cost;
    
    if (!input || (shot.cost < 0.0)) {
        fprintf(stderr, "Unable to read shot cost for type %d.\n", type);
        
        free(shot.name);
        input.close();
        return false;
    }
    
    input >> shot.reloadTime;
    
    if (!input || (shot.reloadTime < 0.0)) {
        fprintf(stderr, "Unable to read shot reload time for type %d.\n", type);
        
        free(shot.name);
        input.close();
        return false;
    }
    
    input >> shot.damage;
    
    if (!input || (shot.damage < 0.0)) {
        fprintf(stderr, "Unable to read shot damage for type %d.\n", type);
        
        free(shot.name);
        input.close();
        return false;
    }
    
    input >> shot.speed;
    
    if (!input || (shot.speed < 0.0)) {
        fprintf(stderr, "Unable to read shot speed for type %d.\n", type);
        
        free(shot.name);
        input.close();
        return false;
    }
    
    input >> buffer[0];
    
    switch (buffer[0]) {
        case 'F': // Falling
            shot.behaviour = Falling; break;
        case 'A': // Aimed
            shot.behaviour = Aimed; break;
        case 'H': // Homing
            shot.behaviour = Homing; break;
        default:
            fprintf(stderr, "Unknown shot behaviour '%c' for type %d.\n", buffer[0], type);
            
            free(shot.name);
            input.close();
            return false;
    }
    
    if (!input) {
        fprintf(stderr, "Unable to read shot behaviour for type %d.\n", type);
        
        free(shot.name);
        input.close();
        return false;
    }
    
    return true;
}

void unloadShot(Shot &shot) {
    if (shot.name) {
        free(shot.name);
        shot.name = NULL;
    }
    
    shot.graphic.DeleteAll();
    
    if (shot.sound) {
        closeSoundFile(shot.sound);
        shot.sound = NULL;
    }
}

bool loadLevel(int index, Level &level) {
    char *temp;
    char buffer[20];
    
    changeToRelativeDirectory(levelsDir);
    
    snprintf(buffer, 20, "%d", index);
    ifstream input(buffer);
    
    if (!input) {
        fprintf(stderr, "Unable to find/open level %d.\n", index);
        
        return false;
    }
    
    level.name = readQuotedString(input);
    
    if (!level.name) {
        fprintf(stderr, "Unable to read name of level %d.\n", index);
        
        input.close();
        return false;
    }
    
    level.introText = readQuotedString(input);

    if (!level.introText) {
        fprintf(stderr, "Unable to read intro text of level %d.\n", index);
        
        free(level.name);
        input.close();
        return false;
    }
    
    changeToRelativeDirectory(musicDir);
    temp = readQuotedString(input);
    
    if (!temp) {
        fprintf(stderr, "Unable to read music file name for level %d.\n", index);

        free(level.name);
        free(level.introText);
        input.close();
        return false;
    }
    
    if (strlen(temp) > 0) {
        level.backgroundMusic = loadSoundFile(temp);
        
        if (!level.backgroundMusic) {
            fprintf(stderr, "Unable to load music for level %d.\n", index);

            free(level.name);
            free(level.introText);
            free(temp);
            input.close();
            return false;
        }
    } else {
        level.backgroundMusic = NULL;
    }
    
    free(temp);
    
    input >> level.columnCount;
    
    if ((level.columnCount <= 0) || !input) {
        fprintf(stderr, "Unable to read column count (or it is invalid) for level %d.\n", index);

        free(level.name);
        free(level.introText);
        
        if (level.backgroundMusic) {
            closeSoundFile(level.backgroundMusic);
        }
        
        input.close();
        return false;
    }
    
    input >> level.rowCount;
    
    if ((level.rowCount <= 0) || !input) {
        fprintf(stderr, "Unable to read row count (or it is invalid) for level %d.\n", index);

        free(level.name);
        free(level.introText);
        
        if (level.backgroundMusic) {
            closeSoundFile(level.backgroundMusic);
        }
        
        input.close();
        return false;
    }
    
    //level.types = (Enemy*)malloc(sizeof(Enemy) * level.rowCount);
    //memset(level.types, '\0', sizeof(Enemy) * level.rowCount);
    
    level.types = new Enemy[level.rowCount];
    
    for (int i = 0; i < level.rowCount; ++i) {
        temp = readQuotedString(input);
        
        if (!temp || !input || !loadEnemy(temp, level.types[i])) {
            fprintf(stderr, "Unable to read enemy information for row %d of level %d.\n", i, index);

            free(level.name);
            free(level.introText);
            
            if (level.backgroundMusic) {
                closeSoundFile(level.backgroundMusic);
            }
        
            free(level.types);
            input.close();
            return false;
        }
    }
    
    input >> level.horizontalSpacing >> level.verticalSpacing;

    if (!input) {
        fprintf(stderr, "Unable to read grid spacing information of level %d.\n", index);
        
        free(level.name);
        free(level.introText);
        
        if (level.backgroundMusic) {
            closeSoundFile(level.backgroundMusic);
        }
        
        free(level.types);
        input.close();
        return false;
    }
    
    input >> level.speed;
    
    if (!input) {
        fprintf(stderr, "Unable to read speed information of level %d.\n", index);
        
        free(level.name);
        free(level.introText);
        
        if (level.backgroundMusic) {
            closeSoundFile(level.backgroundMusic);
        }
        
        free(level.types);
        input.close();
        return false;
    }
    
    input >> level.timeLimited;
    
    if (level.timeLimited) {
        input >> level.timeLimit;
    } else {
        level.timeLimit = 0;
    }
    
    input >> level.bonusModified;
    
    if (level.bonusModified) {
        input >> level.bonusModifierKickIn >> level.bonusModifier;
        
        if (!input || (level.bonusModifierKickIn < 0) || (level.bonusModifier <= 0)) {
            fprintf(stderr, "Unable to read bonus modifier and/or modifier kick in for level %d.\n", index);

            free(level.name);
            free(level.introText);
            
            if (level.backgroundMusic) {
                closeSoundFile(level.backgroundMusic);
            }
        
            free(level.types);
            input.close();
            return false;
        }
    } else {
        level.bonusModifierKickIn = 0;
        level.bonusModifier = 0.0;
    }
    
    input >> level.bonus;
    
    level.endText = readQuotedString(input);
    
    if (!level.endText) {
        fprintf(stderr, "Unable to read end text for level %d.\n", index);

        free(level.name);
        free(level.introText);
        
        if (level.backgroundMusic) {
            closeSoundFile(level.backgroundMusic);
        }
        
        free(level.types);
        input.close();
        return false;
    }
    
    input.close();
    
    return true;
}

void unloadLevel(Level &level) {
    if (level.name) {
        free(level.name);
        level.name = NULL;
    }
    
    if (level.introText) {
        free(level.introText);
        level.introText = NULL;
    }
    
    if (level.endText) {
        free(level.endText);
        level.endText = NULL;
    }
    
    if (level.backgroundMusic) {
        closeSoundFile(level.backgroundMusic);
        level.backgroundMusic = NULL;
    }
    
    if (level.types) {
        for (int i = 0; i < level.rowCount; ++i) {
            unloadEnemy(level.types[i]);
        }
        
        delete[] level.types;
        //free(level.types);
        level.types = NULL;
    }
}

bool loadEnemy(const char *name, Enemy &enemy) {
    char *temp;
    
    if (!name) {
        fprintf(stderr, "Warning: No name (NULL) given to loadEnemy.\n");

        return false;
    }
    
    changeToRelativeDirectory(enemyDir);

    ifstream input(name);
    
    if (!input) {
        fprintf(stderr, "Unable to open data file for enemy \"%s\".\n", name);

        return false;
    }
    
    enemy.name = readQuotedString(input);
    
    if (!enemy.name) {
        fprintf(stderr, "Unable to read full name for enemy \"%s\".\n", name);

        input.close();
        return false;
    }
    
    temp = readQuotedString(input);
    
    if (!temp) {
        fprintf(stderr, "Unable to read graphics path for enemy \"%s\".\n", name);

        free(enemy.name);
        input.close();
        return false;
    }
    
    changeToRelativeDirectory(graphicsDir);

    if (!enemy.graphic.LoadFrom(temp)) {
        fprintf(stderr, "Unable to load graphics for enemy \"%s\".\n", name);

        free(enemy.name);
        free(temp);
        input.close();
        return false;
    }
    
    free(temp);
    
    temp = readQuotedString(input);
    
    if (!temp) {
        fprintf(stderr, "Unable to read hit sound name for enemy \"%s\".\n", name);
        
        free(enemy.name);
        input.close();
        return false;
    }
    
    changeToRelativeDirectory(audioDir);
    
    if (strlen(temp) > 0) {
        if (!(enemy.hitSound = loadSoundFile(temp))) {
            fprintf(stderr, "Unable to load hit sound for enemy \"%s\".\n", name);
            
            free(enemy.name);
            free(temp);
            input.close();
            return false;
        }
    } else {
        enemy.hitSound = NULL;
    }
    
    free(temp);
    
    temp = readQuotedString(input);
    
    if (!temp) {
        fprintf(stderr, "Unable to read die sound name for enemy \"%s\".\n", name);
        
        free(enemy.name);
        input.close();
        return false;
    }
        
    if (strlen(temp) > 0) {
        if (!(enemy.dieSound = loadSoundFile(temp))) {
            fprintf(stderr, "Unable to load die sound for enemy \"%s\".\n", name);
            
            free(enemy.name);
            free(temp);
            input.close();
            return false;
        }
    } else {
        enemy.dieSound = NULL;
    }
    
    free(temp);
    
    input >> enemy.value >> enemy.hitPoints;
    
    if (!input || (enemy.hitPoints <= 0)) {
        fprintf(stderr, "Unable to read value and/or hit points, or hit points were invalid, for enemy \"%s\".\n", name);

        free(enemy.name);
        input.close();
        return false;
    }
    
    input >> enemy.numberOfShotTypes;
    
    if (!input || (enemy.numberOfShotTypes < 0)) {
        fprintf(stderr, "Unable to read number of shot types for enemy \"%s\".\n", name);
        
        free(enemy.name);
        input.close();
        return false;
    }
    
    if (enemy.numberOfShotTypes > 0) {
        enemy.shotTypes = new int[enemy.numberOfShotTypes];
        enemy.shotProbabilities = new double[enemy.numberOfShotTypes];
        
        for (int i = 0; i < enemy.numberOfShotTypes; ++i) {
            input >> enemy.shotTypes[i] >> enemy.shotProbabilities[i];
            
            if (!input) {
                fprintf(stderr, "Unable to read information for shot #%d for enemy \"%s\".\n", i, name);
                
                free(enemy.name);
                free(enemy.shotTypes);
                free(enemy.shotProbabilities);
                input.close();
                return false;
            }
        }
    } else {
        enemy.shotTypes = NULL;
        enemy.shotProbabilities = NULL;
    }
    
    input.close();
    
    return true;
}

void unloadEnemy(Enemy &enemy) {
    if (enemy.name) {
        free(enemy.name);
        enemy.name = NULL;
    }
    
    enemy.graphic.DeleteAll();
    
    if (enemy.shotTypes) {
        delete[] enemy.shotTypes;
        //free(enemy.shotTypes);
        enemy.shotTypes = NULL;
    }
    
    if (enemy.shotProbabilities) {
        delete[] enemy.shotProbabilities;
        //free(enemy.shotProbabilities);
        enemy.shotProbabilities = NULL;
    }
    
    if (enemy.hitSound) {
        closeSoundFile(enemy.hitSound);
        enemy.hitSound = NULL;
    }
    
    if (enemy.dieSound) {
        closeSoundFile(enemy.dieSound);
        enemy.dieSound = NULL;
    }
}
